/*
 * Original theme modified by Evan Rosson
 * http://erosson.com/migrate
 * ---
 *
 * Theme Name: Almodovar
 * Theme URI: http://blog.ratterobert.com/archiv/2005/03/09/almodovar/
 * Description: Das Theme basiert im Ursprung auf Michael Heilemanns <a href="http://binarybonsai.com/kubrick/">Kubrick</a>-Template und ist von dem einen oder anderen Gimmick anderer sehr guter Templates inspiriert worden.
 * Version: 0.7
 * Author: ratte / robert
 * Author URI: http://blog.ratterobert.com/
 * */

/* Begin Typography & Colors */
body {
    font-size: 75%;
    font-family: 'Lucida Grande', 'Trebuchet MS', 'Bitstream Vera Sans', Sans-Serif;
    background-color: #CCF;
    color: #333;
    text-align: center;
}

#page {
    background-color: #fff;
    border: 1px solid #88f;
    text-align: left;
}

#content {
    font-size: 1.2em;
    margin: 1em;
    }

#content p,
#content ul,
#content blockquote {
  line-height: 1.6em;
}

#footer {
    border-top: 1px solid #006;
  margin-top: 2em;
}

small {
    font-family: 'Trebuchet MS', Arial, Helvetica, Sans-Serif;
    font-size: 0.9em;
    line-height: 1.5em;
    }

h1, h2, h3 {
    font-family: 'Trebuchet MS', 'Lucida Grande', Verdana, Arial, Sans-Serif;
    font-weight: bold;
  margin-top: .7em;
  margin-bottom: .7em;
    }

h1 {
  font-size: 2.5em;
}
h2 {
    font-size: 2em;
}
h3 {
    font-size: 1.5em;
}

h1, h2, h3 {
    color: #33a;
}

h1 a, h2 a, h3 a {
    color: #33a;
}

h1, h1 a, h1 a:hover, h1 a:visited,
h2, h2 a, h2 a:hover, h2 a:visited,
h3, h3 a, h3 a:hover, h3 a:visited,
cite {
    text-decoration: none;
}

#content p a:visited {
    color: #004099;
    /*font-weight: normal;*/
}

small, blockquote, strike {
    color: #33a;
}

#links ul ul li, #links li {
    list-style: none;
}

code {
    font: 1.1em 'Courier', 'Courier New', Fixed;
}

acronym, abbr, span.caps {
    font-size: 0.9em;
    letter-spacing: .07em;
}

a {
    color: #0050FF;
    /*text-decoration: none;*/
    text-decoration:underline;
    /*font-weight: bold;*/
}
a:hover {
    color: #0080FF;
}

/* Special case doc-title */
h1.doc-title {
  text-transform: lowercase;
  font-size: 4em;
  margin: 0;
}
h1.doc-title a {
  display: block;
  padding-left: 0.8em;
  padding-bottom: .5em;
  padding-top: .5em;
  margin: 0;
  border-bottom: 1px #fff solid;
}
h1.doc-title,
h1.doc-title a,
h1.doc-title a:visited,
h1.doc-title a:hover {
    text-decoration: none;
    color: #0050FF;
}
/* End Typography & Colors */


/* Begin Structure */
body {
    margin: 0;
    padding: 0;
}

#page {
    background-color: white;
    margin: 0 auto 0 9em;
    padding: 0;
    max-width: 60em;
    border: 1px solid #555596;
}
* html #page {
*   width: 60em;
*   }
*
*   #content {
*       margin: 0 1em 0 3em;
*       }
*
*       #content h1 {
*         margin-left: 0;
*         }
*
*         #footer {
*           padding: 0 0 0 1px;
*               margin: 0;
*                 margin-top: 1.5em;
*                   clear: both;
*                       }
*
*                       #footer p {
*                           margin: 1em;
*                               }
*
*/* End Structure */



/*  Begin Headers */
.description {
    text-align: center;
    }

/* End Headers */


/* Begin Form Elements */
#searchform {
    margin: 1em auto;
    text-align: right;
    }

#searchform #s {
    width: 100px;
    padding: 2px;
    }

#searchsubmit {
    padding: 1px;
    }
/* End Form Elements */


/* Begin Various Tags & Classes */
acronym, abbr, span.caps {
    cursor: help;
}

acronym, abbr {
    border-bottom: 1px dashed #999;
}

blockquote {
    margin: 15px 30px 0 10px;
    padding-left: 20px;
    border-left: 5px solid #CCC;
}

blockquote cite {
    margin: 5px 0 0;
    display: block;
}

hr {
    display: none;
}

a img {
    border: none;
}

.navigation {
    display: block;
    text-align: center;
    margin-top: 10px;
    margin-bottom: 60px;
}
/* End Various Tags & Classes*/

span a { color: #CCC; }

span a:hover { color: #0050FF; }

#navcontainer {
    margin-top: 0px;
    padding-top: 0px;
    width: 100%;
    background-color: #AAF;
    text-align: right;
}

#navlist ul {
    margin-left: 0;
    margin-right: 5px;
    padding-left: 0;
    white-space: nowrap;
}

#navlist li {
    display: inline;
    list-style-type: none;
}

#navlist a {
  padding: 3px 10px;
  color: #fff;
    background-color: #339;
    text-decoration: none;
    border: 1px solid #44F;
  font-weight: normal;
}

#navlist a:hover {
    color: #000;
    background-color: #FFF;
    text-decoration: none;
  font-weight: normal;
}

#navlist a:active, #navlist a.selected {
    padding: 3px 10px;
    color: #000;
    background-color: #EEF;
    text-decoration: none;
    border: 1px solid #CCF;
    font-weight: normal;
}

